<?php

namespace Drupal\search_api_target_bundle_filter\Plugin\search_api\processor;

use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\search_api\IndexInterface;
use Drupal\search_api\Plugin\PluginFormTrait;
use Drupal\search_api\Processor\ProcessorPluginBase;
use Drupal\search_api\SearchApiException;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Allows you to filter a reference field's value by the bundle of the target entities.
 *
 * @SearchApiProcessor(
 *   id = "target_bundle_filter",
 *   label = @Translation("Target Bundle Filter"),
 *   description = @Translation("Provides a filter such that values from only selected target bundles of a field will be included in the results."),
 *   stages = {
 *     "preprocess_index" = -45
 *   }
 * )
 */
class TargetBundleFilter extends ProcessorPluginBase implements PluginFormInterface {

  use PluginFormTrait;

  /**
   * Static cache for getReferenceFields() return values, keyed by index ID.
   *
   * @var string[][][]
   *
   * @see \Drupal\search_api_target_bundle_filter\Plugin\search_api\processor\TargetBundleFilter::getReferenceFields()
   */
  protected static $indexReferenceFields = [];

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface|null
   */
  protected $entityTypeManager;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface |null
   */
  protected $entityFieldManager;

  protected $entityTypeBundleInfo;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    /** @var static $processor */
    $processor = parent::create($container, $configuration, $plugin_id, $plugin_definition);

    $processor->setEntityTypeManager($container->get('entity_type.manager'));
    $processor->setEntityTypeBundleInfo($container->get('entity_type.bundle.info'));
    $processor->setEntityFieldManager($container->get('entity_field.manager'));

    return $processor;
  }

  /**
   * Retrieves the entity type manager service.
   *
   * @return \Drupal\Core\Entity\EntityTypeManagerInterface
   *   The entity type manager service.
   */
  public function getEntityTypeManager() {
    return $this->entityTypeManager ?: \Drupal::entityTypeManager();
  }

  /**
   * Sets the entity type manager service.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   *
   * @return $this
   */
  public function setEntityTypeManager(EntityTypeManagerInterface $entity_type_manager) {
    $this->entityTypeManager = $entity_type_manager;
    return $this;
  }

  /**
   * Retrieves the entity display repository.
   *
   * @return \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   *   The entity entity display repository.
   */
  public function getEntityTypeBundleInfo() {
    return $this->entityTypeBundleInfo ?: \Drupal::service('entity_type.bundle.info');
  }

  /**
   * Sets the entity type bundle info.
   *
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The new entity type bundle info.
   *
   * @return $this
   */
  public function setEntityTypeBundleInfo(EntityTypeBundleInfoInterface $entity_type_bundle_info) {
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
    return $this;
  }
  /**
   * Retrieves the entity field manager.
   *
   * @return \Drupal\Core\Entity\EntityFieldManagerInterface
   *   The entity field manager.
   */
  public function getEntityFieldManager() {
    return $this->entityFieldManager ?: \Drupal::service('entity_field.manager');
  }

  /**
   * Sets the entity field manager.
   *
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The new entity field manager.
   *
   * @return $this
   */
  public function setEntityFieldManager(EntityFieldManagerInterface $entity_field_manager) {
    $this->entityFieldManager = $entity_field_manager;
    return $this;
  }


  /**
   * {@inheritdoc}
   */
  public static function supportsIndex(IndexInterface $index) {
    $processor = new static(['#index' => $index], 'target_bundle_filter', []);
    return (bool) $processor->getReferenceFields();
  }

  /**
   * Finds all reference fields for this processor's index.
   *
   * Fields are returned if:
   * - They point to an entity type
   * - They reference more than one bundle
   *
   * @return string[][]
   *   An array containing all fields of the index for which target data
   *   might be retrievable. The keys are those field's IDs, the values are
   *   associative arrays containing the allowed bundles of those fields.
   */
  protected function getReferenceFields() {
    if (!isset(static::$indexReferenceFields[$this->index->id()])) {
      $bundle_info = $this->getEntityTypeBundleInfo()->getAllBundleInfo();
      $field_options = [];
      foreach ($this->index->getFields() as $field_id => $field) {
        try {
          $property = $field->getDataDefinition();
          $datasource = $field->getDatasource();
        }
        catch (SearchApiException $e) {
          // Will probably cause other problems elsewhere, but here we can just
          // ignore it.
          continue;
        }
        if (!$datasource) {
          continue;
        }
        if ($property->getDataType() !== 'field_item:entity_reference') {
          continue;
        }

        $entity_type_id = $datasource->getEntityTypeId();
        // Field instances per bundle can reference different vocabularies, make
        // sure we add them all.
        $bundles = $bundle_info[$entity_type_id];
        // For every bundle
        foreach ($bundles as $bundle_id => $bundle) {
          $bundle_fields = $this->getEntityFieldManager()
            ->getFieldDefinitions($entity_type_id, $bundle_id);
          // Check if this bundle has the taxonomy entity reference field.

          if (array_key_exists($property->getFieldDefinition()->getName(), $bundle_fields)) {
            /** @var \Drupal\Core\Field\FieldDefinitionInterface $field_definition */
            $field_definition = $bundle_fields[$property->getFieldDefinition()->getName()];
            $bundle_settings = $field_definition->getSettings();
            if (!empty($bundle_settings['handler_settings']['target_bundles'])
              && !empty($bundle_settings['target_type']) ) {
              $field_options += [$field_id => []];
              // Convert to labels
              $options = $bundle_settings['handler_settings']['target_bundles'];
              foreach ($options as $key => $option) {
                $label = $bundle_info[$bundle_settings['target_type']][$key]['label'];
                $options[$key] = $label;
              }
              $field_options[$field_id] += $options;
            }
            else {
              $field_options += [$field_id => []];
            }
          }
        }
      }
      // Filter out fields with only one bundle
      foreach ($field_options as $field_id => $options) {
        if (count($options) <= 1) {
          unset($field_options[$field_id]);
        }
      }

      static::$indexReferenceFields[$this->index->id()] = $field_options;
    }

    return static::$indexReferenceFields[$this->index->id()];
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'convert_to_labels' => 0,
      'fields' => [],
    ];
  }


  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $formState) {
    $form['#description'] = $this->t('Select fields to be filtered by reference target bundle.');
    $form['convert_to_labels'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Convert field values from id to label?'),
      '#default_value' => $this->configuration['convert_to_labels'] == 1,
      '#return_value' => 'label',
    ];

    foreach ($this->getReferenceFields() as $field_id => $options) {
      $enabled = !empty($this->configuration['fields'][$field_id]);
      $name = "processors[target_bundle_filter][settings][fields][{$field_id}][field_status]";
      $form['fields'][$field_id]['field_status'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Enable %field', ['%field' => $this->index->getField($field_id)->getLabel()]),
        '#default_value' => $enabled ? 'on' : null,
        '#return_value' => 'on',
      ];

      $form['fields'][$field_id]['selection_type'] = [
        '#type' => 'radios',
        '#title' => $this->t('Which bundles should be indexed?'),
        '#options' => [
          'only' => $this->t('Only those selected'),
          'all_except' => $this->t('All except those selected'),
        ],
        '#access' => count($options) > 0,
        '#states' => [
          // Show this checkbox only if the field is enabled above
          'visible' => [
            ':input[name="' . $name . '"]' => ['checked' => TRUE],
          ],
        ],
        '#default_value' => $this->configuration['fields'][$field_id]['selection_type'] ?? [],
      ];
      reset($options);
      $form['fields'][$field_id]['selected'] = [
        '#type' => 'checkboxes',
        '#title' => $this->t('Select bundles'),
        '#description' => $this->t("Please pick the bundles that should be indexed. "),
        '#options' => $options,
        '#default_value' => $enabled ? $this->configuration['fields'][$field_id]['selected'] : [],
        '#size' => min(4, count($options)),
        '#multiple' => TRUE,
        '#access' => count($options) > 0,
        '#states' => [
          // Show this checkbox only if the field is enabled above
          'visible' => [
            ':input[name="' . $name . '"]' => ['checked' => TRUE],
          ],
        ],
      ];
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $formState) {
    $fields = [];
    foreach ($formState->getValue('fields', []) as $field_id => $values) {
      if (!empty($values['field_status'])) {
        if (empty($values['selection_type'])) {
          $formState->setError($form['fields'][$field_id]['selection_type'], $this->t('You need to select how to filter bundles for each enabled field.'));
        }
        else {
          $fields[$field_id] = [
            'field_status' => $values['field_status'],
            'selection_type' => $values['selection_type'],
            'selected' => array_filter($values['selected'], function ($a) {
              return $a !== 0;
            }),
          ];
        }
      }
    }
    $formState->setValue('fields', $fields);
    $convert_to_labels = $formState->getValue('convert_to_labels', []);
    $formState->setValue('convert_to_labels', $convert_to_labels == 'label' ? 1 : 0);
    if (!$fields) {
      $formState->setError($form['fields'], $this->t('You need to enable target filter on at least one field or disable the processor.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function preprocessIndexItems(array $items) {
    $convert_labels = $this->configuration['fields'] == 'label';

    /** @var \Drupal\search_api\Item\ItemInterface $item */
    foreach ($items as $item) {
      try {
        $object = $item->getOriginalObject();
      } catch (SearchApiException $e) {
        continue;
      }

      foreach ($this->configuration['fields'] as $field_id => $field_config) {
        if ($field_config['field_status'] !== "on") {
          continue;
        }
        /** @var \Drupal\search_api\Item\FieldInterface $field */
        $field = $item->getField($field_id);
        if (!$field) {
          continue;
        }
        // Do we want "only" certain bundles or "all except"
        $want_in_array = $field_config['selection_type'] == "only";

        $path = $field->getPropertyPath();
        $original_field = $object->get($path); // TODO: Validate path
        $references = $original_field->referencedEntities();

        // Filter $references down to wanted bundles and index by id
        $ids = [];
        foreach ($references as $reference) {
          if(array_key_exists($reference->bundle(), $field_config['selected']) === $want_in_array) {
            $ids[$reference->id()] = $reference;
          }
        }
        // Check each field value against filtered values
        $values = count($ids) ? $field->getValues() : [];
        foreach ($values as $delta => $entity_id) {
          if (array_key_exists($entity_id, $ids) === FALSE) {
            unset($values[$delta]);
          }
          else {
            if ($convert_labels) {
              $values[$delta] = ($ids[$entity_id])->label();
            }
          }
        }
        $field->setValues($values);
      }
    }
  }

}
